@@ -14,7 +14,7 @@ from datetime import datetime
 from functools import lru_cache
 from glob import glob
 from itertools import groupby
-from typing import Iterable, Set
+from typing import Iterable, List, Optional, Set
 
 from b2luigi.basf2_helper.utils import get_basf2_git_hash
 from b2luigi.batch.processes import BatchProcess, JobStatus
@@ -22,6 +22,7 @@ from b2luigi.core.settings import get_setting
 from b2luigi.core.utils import flatten_to_dict, get_log_file_dir, get_task_file_dir
 from jinja2 import Template
 from luigi.target import Target
+from retry import retry
 
 
 class Gbasf2Process(BatchProcess):
@@ -656,7 +657,7 @@ class Gbasf2Process(BatchProcess):
             os.makedirs(tmp_output_dir_path, exist_ok=True)
 
             # Need a set files to repeat download for FAILED ones only
-            monitoring_failed_downloads_file = os.path.join(tmp_output_dir_path, "failed_files.txt")
+            monitoring_failed_downloads_file = os.path.abspath(os.path.join(tmp_output_dir_path, "failed_files.txt"))
             monitoring_download_file_stem, monitoring_downloads_file_ext = os.path.splitext(monitoring_failed_downloads_file)
             old_monitoring_failed_downloads_file = f"{monitoring_download_file_stem}_old{monitoring_downloads_file_ext}"
 
@@ -665,7 +666,8 @@ class Gbasf2Process(BatchProcess):
 
                 ds_get_command = shlex.split(f"gb2_ds_get --force {dataset_query_string} "
                                              f"--failed_lfns {monitoring_failed_downloads_file}")
-                print("Downloading dataset with command ", " ".join(ds_get_command))
+                print(f"Downloading dataset into\n  {tmp_output_dir_path}\n  with command\n  ",
+                      " ".join(ds_get_command))
 
             # Any further time is based on the list of files from failed downloads
             else:
@@ -674,7 +676,8 @@ class Gbasf2Process(BatchProcess):
                 ds_get_command = shlex.split(f"gb2_ds_get --force {dataset_query_string} "
                                              f"--input_dslist {old_monitoring_failed_downloads_file} "
                                              f"--failed_lfns {monitoring_failed_downloads_file}")
-                print("Downloading remaining files from dataset with command ", " ".join(ds_get_command))
+                print(f"Downloading remaining files into\n  {tmp_output_dir_path}\n  with command\n  ",
+                      " ".join(ds_get_command))
 
             stdout = run_with_gbasf2(ds_get_command, cwd=tmp_output_dir_path, capture_output=True).stdout
             print(stdout)
@@ -695,24 +698,20 @@ class Gbasf2Process(BatchProcess):
                 if os.path.isfile(monitoring_failed_downloads_file):
                     os.remove(monitoring_failed_downloads_file)
 
-            tmp_output_dir = os.path.join(tmp_output_dir_path, self.gbasf2_project_name)
+            tmp_project_dir = os.path.join(tmp_output_dir_path, self.gbasf2_project_name)
             if not self._local_gb2_dataset_is_complete(output_file_name, check_temp_dir=True):
                 raise RuntimeError(
-                    f"Download incomplete. The downloaded set of files in {tmp_output_dir} is not equal to the " +
+                    f"Download incomplete. The downloaded set of files in {tmp_project_dir} is not equal to the " +
                     f"list of dataset files on the grid for project {self.gbasf2_project_name}.",
                 )
 
             print(f"Download of {self.gbasf2_project_name} files successful.\n"
                   f"Moving output files to directory: {output_dir_path}")
-            if os.path.exists(output_dir_path):
-                shutil.rmtree(output_dir_path)
             # sub00 and other sub<xy> directories in case of other datasets
-            sub_directories = glob(os.path.join(tmp_output_dir, "sub*"))
-            for sub_dir in sub_directories:
-                if not os.path.isdir(sub_dir):
-                    raise NotADirectoryError(f"{sub_dir} is not a directory.")
-                shutil.move(src=sub_dir, dst=output_dir_path)
-            shutil.rmtree(tmp_output_dir_path)
+            _move_downloaded_dataset_to_output_dir(
+                project_download_path=tmp_project_dir,
+                output_path=output_dir_path
+            )
 
     def _download_logs(self):
         """
@@ -789,7 +788,11 @@ def check_dataset_exists_on_grid(gbasf2_project_name, dirac_user=None):
     return len(lpns) > 0
 
 
-def get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user=None, n_retries=5):
+@retry(
+    (json.decoder.JSONDecodeError, subprocess.CalledProcessError),
+    tries=4, delay=2, backoff=3  # retry after 2,6,18,108 s
+)
+def get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user=None):
     """
     Returns a dictionary for all jobs in the project with a structure like the
     following, which I have taken and adapted from an example output::
@@ -827,18 +830,7 @@ def get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user=None, n_r
             "Probably there was an error during the project submission when running the gbasf2 command.\n"
         )
     job_status_json_string = proc.stdout
-    try:
-        job_status_dict = json.loads(job_status_json_string)
-        return job_status_dict
-    except json.decoder.JSONDecodeError as err:
-        warnings.warn(
-            " Failed to decode job status json string: \n" +
-            job_status_json_string
-        )
-        if n_retries > 0:
-            print(f"Trying to get new job status ({n_retries - 1} retries left)")
-            return get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user, n_retries=n_retries - 1)
-        raise err
+    return json.loads(job_status_json_string)
 
 
 def check_project_exists(gbasf2_project_name, dirac_user=None):
@@ -904,7 +896,7 @@ def get_gbasf2_env(gbasf2_install_directory=None):
             os.path.join("/cvmfs/belle.kek.jp/grid/gbasf2/pro", os.pardir, os.pardir)
         )
         gbasf2_install_directory = get_setting("gbasf2_install_directory", default=default_gbasf2_install_directory)
-    gbasf2_setup_path = os.path.join(gbasf2_install_directory, "BelleDIRAC/gbasf2/tools/setup")
+    gbasf2_setup_path = os.path.join(gbasf2_install_directory, "BelleDIRAC/gbasf2/tools/setup.sh")
     if not os.path.isfile(gbasf2_setup_path):
         raise FileNotFoundError(
             f"Could not find gbasf2 setup file at:\n{gbasf2_setup_path}.\n"
@@ -922,7 +914,10 @@ def get_gbasf2_env(gbasf2_install_directory=None):
     # The gbasf2 setup script on sets HOME to /ext/home/ueda if it's unset,
     # which later causes problems in the gb2_proxy_init subprocess. Therefore,
     # reset it to the caller's HOME.
-    gbasf2_env["HOME"] = os.getenv("HOME")
+    try:
+        gbasf2_env["HOME"] = os.environ["HOME"]
+    except KeyError:
+        pass
     return gbasf2_env
 
 
@@ -933,7 +928,8 @@ def get_proxy_info():
         "gbasf2_utils/gbasf2_proxy_info.py"
     )
 
-    # Setting ``initalize_proxy=False`` is vital here, otherwise we get an infinite loop
+    # Setting ``ensure_proxy_initialized=False`` is vital here, otherwise we get
+    # an infinite loop because run_with_gbasf2 will then check for the proxy info
     proc = run_with_gbasf2(
         [proxy_info_script_path],
         capture_output=True,
@@ -942,8 +938,12 @@ def get_proxy_info():
     return json.loads(proc.stdout)
 
 
+@lru_cache(maxsize=None)
 def get_dirac_user():
     """Get dirac user name."""
+    # ensure proxy is initialized, because get_proxy_info can't do it, otherwise
+    # it causes an infinite loop
+    setup_dirac_proxy()
     try:
         return get_proxy_info()["username"]
     except KeyError as err:
@@ -990,13 +990,12 @@ def setup_dirac_proxy():
                 cmd=proxy_init_cmd,
                 output=("There seems to be an error in the output of gb2_proxy_init."
                         f"\nCommand output:\n{out}"),
-                stderr=()
             )
         # if no  wrong password and no other errors, stop loop and return
         return
 
 
-def query_lpns(ds_query, dirac_user=None):
+def query_lpns(ds_query: str, dirac_user: Optional[str] = None) -> List[str]:
     """
     Query DIRAC for LPNs matching query, and return them as a list.
 
@@ -1009,13 +1008,20 @@ def query_lpns(ds_query, dirac_user=None):
         os.path.dirname(os.path.realpath(__file__)), "gbasf2_utils/gbasf2_ds_list.py"
     )
 
+    query_cmd = [ds_list_script_path, "--dataset", ds_query, "--user", dirac_user]
     proc = run_with_gbasf2(
-        [ds_list_script_path, "--dataset", ds_query, "--user", dirac_user],
+        query_cmd,
         capture_output=True,
         ensure_proxy_initialized=True,
     )
 
-    return json.loads(proc.stdout)
+    lpns = json.loads(proc.stdout)
+    if not isinstance(lpns, list):
+        raise TypeError(
+            f"Expected gbasf2 dataset query to return list, but instead the command\n{query_cmd}\n" +
+            f"returned: {lpns}"
+        )
+    return lpns
 
 
 def get_unique_project_name(task):
@@ -1102,8 +1108,49 @@ def get_unique_lfns(lfns: Iterable[str]) -> Set[str]:
     # if dataset does not follow gbasf2 v5 convention, assume it was produced
     # with old release and does not contain duplicates
     if not all(lfn_follows_gb2v5_convention(lfn) for lfn in lfns):
-        return list(lfns)
+        return set(lfns)
     # if it is of the gbasf v5 form, group the outputs by the substring upto the
     # reschedule number and return list of maximums of each group
     lfns = sorted(lfns, key=_get_lfn_upto_reschedule_number)
     return {max(lfn_group) for _, lfn_group in groupby(lfns, key=_get_lfn_upto_reschedule_number)}
+
+
+def _move_downloaded_dataset_to_output_dir(project_download_path: str, output_path: str) -> None:
+    """
+    Move files downloaded downloaded to the grid to their final output location.
+
+    In the ``Gbasf2Process``, the outputs are usually downloaded with ``gb2_ds_get`` to
+    a temporary directory (``project_download_path``)  with a structure like
+
+        <result_dir>/B.root.partial/<project_name>
+        ├── sub00/job_name*B.root
+        ├── sub01/job_name*B.root
+        ├── …
+
+    This function moves those files to their final ``output_path`` directory which has
+    the same name as the original root file (e.g. ``B.root``) to fullfill the luigi
+    output definition. This output directory has the structure
+
+        <result_dir>/B.root/job_name*B.root
+
+    :param project_download_path: Directory into which ``gb2_ds_get`` downloaded the
+        grid dataset. The contents should be ``sub<xy>`` datablocks containing root files.
+    :param output_path: Final output directory into which the ROOT files should be copied.
+    """
+    # the download shouldn't happen if the output already exists, but assert that's the case just to be sure
+    if os.path.exists(output_path):
+        raise FileExistsError(f"Output directory {output_path} already exists.")
+    sub_directories = glob(os.path.join(project_download_path, "sub*"))
+    try:
+        os.makedirs(output_path, exist_ok=True)
+        for sub_dir in sub_directories:
+            root_fpaths = glob(os.path.join(sub_dir, "*.root"), recursive=False)
+            if not root_fpaths:
+                raise RuntimeError(f"Cannot find any .root files in {sub_dir}.")
+            for root_fpath in root_fpaths:
+                shutil.move(src=root_fpath, dst=output_path)
+    except BaseException:
+        # In case the moving fails (should never happen), remove output_path to make sure that the
+        # task isn't accidentally marked as complete
+        shutil.rmtree(output_path, ignore_errors=True)
+        raise
